import logging
import time
import requests
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import BackendApplicationClient

logger = logging.getLogger(__name__)


class ManagedOAuth2Session(OAuth2Session):
    """
    A drop-in replacement for requests.Session/OAuth2Session
    that transparently handles OAuth2 token acquisition
    and refresh for both client credentials and refresh token flows.
    """
    def __init__(
        self,
        token_url: str,
        client_id: str,
        client_secret: str,
        extra_params: dict = None,
        **kwargs
    ):
        if extra_params and "grant_type" in extra_params:
            raise ValueError(
                "Specifying 'grant_type' in extra_params is not allowed. "
                "Only client credentials grant type is supported."
            )
        client = BackendApplicationClient(client_id=client_id)
        super().__init__(client=client, **kwargs)
        self.token_url = token_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.extra_params = extra_params or {}
        self._fetching_token = False

    def _get_auto_refresh_kwargs(self) -> dict:
        """Prepare kwargs for auto token refresh"""
        return {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            **self.extra_params
        }

    def _acquire_token(self) -> dict:
        """Acquire a new access token using client credentials grant."""
        logger.info("Acquiring new access token via OAuth2.")
        self._fetching_token = True
        try:
            token = super().fetch_token(
                token_url=self.token_url,
                client_id=self.client_id,
                client_secret=self.client_secret,
                include_client_id=True,
                **self.extra_params
            )
            self.token = token
            return token
        finally:
            self._fetching_token = False

    def _reacquire_if_expired(self) -> None:
        """Reacquire token if it is expired or not present."""
        now = time.time()
        expires_at = self.token.get("expires_at", 0) if self.token else 0
        if not self.token or (expires_at and expires_at <= now):
            if not (self.token and self.token.get("refresh_token")):
                # Only reacquire if not already fetching token
                if not self._fetching_token:
                    self._acquire_token()

    def request(self, method, url, **kwargs) -> requests.Response:
        """
        Override request to handle token
        reacquisition and auto-refresh.
        """
        self._reacquire_if_expired()
        if self.token and self.token.get("refresh_token"):
            self.auto_refresh_url = self.token_url
            self.auto_refresh_kwargs = self._get_auto_refresh_kwargs()
        return super().request(method, url, **kwargs)

    @property
    def access_token(self) -> str:
        """Return the current access token."""
        return self.token.get("access_token") if self.token else None

    @property
    def refresh_token_value(self) -> str:
        """Return the current refresh token."""
        return self.token.get("refresh_token") if self.token else None

    @property
    def access_token_expiry(self) -> float:
        """Return the expiry time of the current access token."""
        return self.token.get("expires_at", 0) if self.token else 0



import os
import asyncio
import imaplib
import smtplib
from typing import Sequence
import email as email_module
from email import encoders
from email.utils import formataddr, parseaddr
from email.message import EmailMessage
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

from bs4 import BeautifulSoup


class EmailSender:
    """
    Class to send emails using SMTP server.
    It supports both synchronous and asynchronous sending of emails,
    as well as attachments.
    """
    def __init__(
        self,
        smtp_server: str | None = None,
        smtp_port: int | None = None,
        sender_email: str | None = None,
        sender_password: str | None = None,
        sender_name: str | None = None,
        reply_to_email: str | None = None,
        reply_to_name: str | None = None
    ):
        self._smtp_server = smtp_server or \
            os.environ.get("SMTP_SERVER", "smtp.randers.dk")
        self._smtp_port = smtp_port or int(os.environ.get("SMTP_PORT", 25))

        self.sender_email = sender_email
        self._sender_password = sender_password
        if sender_email:
            if self._is_valid_address(sender_email):
                if sender_name:
                    self.sender = (sender_name, sender_email)
                else:
                    self.sender = sender_email
            else:
                raise ValueError(
                    f"Invalid sender email address: {sender_email}"
                )
        else:
            self.sender = ""

        if reply_to_email:
            if self._is_valid_address(reply_to_email):
                if reply_to_name:
                    self.reply_to = (reply_to_name, reply_to_email)
                else:
                    self.reply_to = reply_to_email
            else:
                raise ValueError(
                    f"Invalid reply-to email address: {reply_to_email}"
                )
        else:
            self.reply_to = ""

        if not self._can_connect():
            raise ConnectionError(
                f"Cannot connect to SMTP server "
                f"{self._smtp_server}:{self._smtp_port}"
            )

    def _is_valid_address(self, address) -> bool:
        """
        Validate an address header.

        Accepts either:
        • a (name, email) tuple, or
        • a plain email string.
        """
        # Case 1: address is a plain string containing an email
        if isinstance(address, str):
            parsed = parseaddr(address)[1]
            return "@" in address and parsed == address

        # Case 2: address is a (name, email) tuple
        if (
            isinstance(address, tuple)
            and len(address) == 2
            and isinstance(address[1], str)
            and "@" in address[1]
            and "@" not in address[0]
            and parseaddr(formataddr(address))[1] == address[1]
        ):
            return True

        return False

    def _can_connect(self) -> bool:
        """
        Method to check if connection to
        SMTP server can be established.
        """
        try:
            with smtplib.SMTP(
                host=self._smtp_server,
                port=self._smtp_port,
                timeout=5
            ) as server:
                server.ehlo()
            return True
        except Exception:
            return False

    def _normalize_addresses(
        self,
        addresses: str | tuple[str, str]
            | Sequence[str | tuple[str, str]] | None,
    ) -> list[str | tuple[str, str]]:
        if addresses is None:
            return []
        if isinstance(addresses, str) and self._is_valid_address(addresses):
            return [addresses]
        elif (
            isinstance(addresses, tuple)
            and self._is_valid_address(addresses)
        ):
            return [addresses]
        else:
            for addr in addresses:
                if not self._is_valid_address(addr):
                    raise ValueError(f"Invalid email address: {addr}")
        return list(addresses)

    def _build_message(
        self,
        sender: str | tuple[str, str],
        reply_to: str | tuple[str, str],
        recipients: str | tuple[str, str] | Sequence[str | tuple[str, str]],
        subject: str,
        body: str,
        cc: str | tuple[str, str] | Sequence[str | tuple[str, str]] | None,
        attachments: Sequence[
            str | tuple[str, bytes | bytearray | memoryview]] | None
    ) -> tuple[MIMEMultipart, str, Sequence[str]]:
        """
        Method to build the message object and return it
        along with from_addr and to_addrs.
        """
        attachments = attachments or []
        cc_list = self._normalize_addresses(cc)
        recipients_list = self._normalize_addresses(recipients)

        to_headers = recipients_list + cc_list

        for addr in [sender] + to_headers + ([reply_to] if reply_to else []):
            if not self._is_valid_address(addr):
                raise ValueError(f"Invalid email address: {addr}")

        msg = MIMEMultipart("mixed")
        msg["From"] = formataddr(sender) if isinstance(sender, tuple) \
            else sender

        if reply_to:
            msg["Reply-To"] = formataddr(reply_to) \
                if isinstance(reply_to, tuple) else reply_to

        if recipients:
            msg["To"] = ", ".join(
                formataddr(addr) if isinstance(addr, tuple) else addr
                for addr in recipients_list
            )

        if cc_list:
            msg["Cc"] = ", ".join(
                formataddr(addr) if isinstance(addr, tuple) else addr
                for addr in cc_list
            )

        msg["Subject"] = subject or ""

        soup = BeautifulSoup(body, "html.parser")

        if soup.find():  # If there are any HTML tags, treat as HTML email
            alt = MIMEMultipart("alternative")
            plain_text = soup.get_text(separator="\n", strip=True)
            alt.attach(MIMEText(plain_text, "plain", "utf-8"))
            alt.attach(MIMEText(body, "html", "utf-8"))
            msg.attach(alt)
        else:
            msg.attach(MIMEText(body or "", "plain", "utf-8"))

        for att in attachments:
            if isinstance(att, str):
                # If value in attachments is a string,
                # treat it as a file path
                with open(att, "rb") as f:
                    filename = os.path.basename(att)
                    content = f.read()
            elif (isinstance(att, tuple)
                    and len(att) == 2
                    and isinstance(att[0], str)
                    and isinstance(att[1], (bytes, bytearray, memoryview))):
                # If value in attachments is a tuple of (filename, content),
                # content may be bytes, bytearray, or memoryview and will be
                # normalized to bytes before attaching.
                filename, content = att[0], bytes(att[1])
            else:
                raise ValueError(
                    "Attachments must be file paths "
                    "or (filename, content) tuples."
                )

            part = MIMEBase("application", "octet-stream")
            part.set_payload(content)
            encoders.encode_base64(part)
            part.add_header(
                "Content-Disposition",
                f'attachment; filename="{filename}"'
            )
            msg.attach(part)
        from_addr = sender[1] if isinstance(sender, tuple) else sender
        to_addrs = [
            addr[1] if isinstance(addr, tuple) else addr
            for addr in to_headers
        ]

        return msg, from_addr, to_addrs

    def send_email(
        self,
        recipients: str | tuple[str, str] | Sequence[str | tuple[str, str]],
        sender: str | tuple[str, str] = "",
        reply_to: str | tuple[str, str] = "",
        subject: str = "",
        body: str = "",
        cc: str | tuple[str, str]
        | Sequence[str | tuple[str, str]] | None = None,
        attachments: Sequence[str | tuple[str, bytes]] | None = None
    ) -> None:
        """
        Sends an email with the specified parameters (sync).
        Will try to authenticate with the SMTP server.
        if sender_email and sender_password were provided in the constructor.
        """
        with smtplib.SMTP(
            host=self._smtp_server,
            port=self._smtp_port,
            timeout=60
        ) as server:
            server.ehlo()
            try:
                server.starttls()
                server.ehlo()
            except smtplib.SMTPException:
                # STARTTLS may be unsupported on some port 25 setups.
                # Fallback to unencrypted connection
                # if STARTTLS fails or is not supported.
                pass
            if self.sender_email and self._sender_password:
                if sender:
                    raise ValueError(
                        "Cannot specify sender when using "
                        "authenticated email sending."
                    )
                else:
                    server.login(
                        user=self.sender_email,
                        password=self._sender_password
                    )

            sender = sender or self.sender or self.sender_email

            if not sender:
                raise ValueError("A sender must be specified.")

            if not recipients and not cc:
                raise ValueError(
                    "At least one recipient (recipients or cc)"
                    " must be specified."
                )

            msg, from_addr, to_addrs = self._build_message(
                sender=sender,
                reply_to=reply_to or self.reply_to,
                recipients=recipients,
                subject=subject,
                body=body,
                cc=cc,
                attachments=attachments
            )

            server.sendmail(
                from_addr=from_addr,
                to_addrs=to_addrs,
                msg=msg.as_string()
            )

    async def send_email_async(
        self,
        sender: str | tuple[str, str] = "",
        reply_to: str | tuple[str, str] | None = None,
        recipients: str | tuple[str, str]
        | Sequence[str | tuple[str, str]] | None = None,
        subject: str = "",
        body: str = "",
        cc: str | tuple[str, str]
        | Sequence[str | tuple[str, str]] | None = None,
        attachments: Sequence[str | tuple[str, bytes]] | None = None
    ) -> None:
        """
        Sends an email with the specified parameters (async).
        Will try to authenticate with the SMTP server.
        if sender_email and sender_password were provided in the constructor.
        """
        import aiosmtplib
        if self.sender_email and self._sender_password and sender:
            raise ValueError(
                "Cannot specify sender when using authenticated email sending."
            )

        sender = sender or self.sender or self.sender_email

        if not sender:
            raise ValueError("A sender must be specified.")

        recipients_list = self._normalize_addresses(recipients)
        if not recipients_list and not cc:
            raise ValueError(
                "At least one recipient (recipients or cc) must be specified."
            )

        msg, from_addr, to_addrs = self._build_message(
            sender=sender,
            reply_to=reply_to or self.reply_to,
            recipients=recipients_list,
            subject=subject,
            body=body,
            cc=cc,
            attachments=attachments
        )

        async with aiosmtplib.SMTP(
            hostname=self._smtp_server,
            port=self._smtp_port,
            timeout=60,
        ) as server:
            await server.ehlo()
            try:
                await server.starttls()
                await server.ehlo()
            except (
                aiosmtplib.errors.SMTPNotSupported,
                aiosmtplib.errors.SMTPException
            ):
                # STARTTLS may be unsupported on some port 25 setups.
                # Fallback to unencrypted connection
                # if STARTTLS fails or is not supported.
                pass

            if self.sender_email and self._sender_password:
                await server.login(self.sender_email, self._sender_password)

            await server.send_message(
                msg,
                sender=from_addr,
                recipients=to_addrs
            )


class EmailReader:
    """
    Class to read emails from an IMAP server.
    """
    def __init__(
        self,
        email: str,
        password: str,
        imap_server: str | None = None,
        imap_port: int | None = None,
    ):
        if not all([email, password]):
            raise ValueError("Email and password must be provided.")
        self._imap_server = imap_server or \
            os.environ.get("IMAP_SERVER", "imap.randers.dk")
        self._imap_port = imap_port or int(os.environ.get("IMAP_PORT", 143))
        self.email = email
        self.password = password
        if not self._can_connect():
            raise ConnectionError(
                f"Cannot connect to IMAP server "
                f"{self._imap_server}:{self._imap_port}"
            )

    def _can_connect(self) -> bool:
        """
        method to check if connection to
        IMAP server can be established.
        """
        try:
            with imaplib.IMAP4(
                host=self._imap_server,
                port=self._imap_port
            ) as server:
                server.starttls()
                msg, _ = server.login(self.email, self.password)
                if msg != "OK":
                    raise ConnectionError("IMAP login failed.")
            return True
        except Exception:
            return False

    def list_mailboxes(self) -> list[str]:
        """
        List all mailboxes in the IMAP account.
        """
        with imaplib.IMAP4(
            host=self._imap_server,
            port=self._imap_port
        ) as server:
            server.starttls()
            server.login(self.email, self.password)
            status, mailboxes = server.list()
            if status != "OK":
                raise ConnectionError("Failed to list mailboxes.")
            return [
                mailbox.decode().split(' "/" ')[-1]
                for mailbox in mailboxes
            ]

    def get_emails(
        self,
        mailbox: str = "INBOX",
        criteria: str = "ALL",
        set_flags: str | None = "\\Seen",
        del_flags: str | None = None,
        max: int | None = None
    ) -> tuple[list[EmailMessage], list[bytes]]:
        """
        Retrieve emails from the specified mailbox
        matching the search criteria.

        param mailbox: The mailbox to search in (e.g., "INBOX")
        param criteria: The IMAP search criteria (e.g., "ALL", "UNSEEN")
        param set_flags: Optional IMAP flags to set on the emails
            after fetching (e.g., "\\Seen" to mark as seen)
        param del_flags: Optional IMAP flags to remove from the emails
            after fetching (e.g., "\\Seen" to mark as unseen)
        param max: Optional maximum number of emails to fetch (all if None)
        """
        with imaplib.IMAP4(
            host=self._imap_server,
            port=self._imap_port
        ) as server:
            server.starttls()
            server.login(self.email, self.password)
            server.select(mailbox)
            status, data = server.search(None, criteria)
            if status != "OK":
                raise ConnectionError("Failed to search emails.")
            all_ids = data[0].split()
            if max is not None:
                email_ids = all_ids[:max]
            else:
                email_ids = all_ids
            emails = []
            failed_to_fetch_ids = []
            for email_id in email_ids:
                status, msg_data = server.fetch(email_id, "(RFC822)")
                if status != "OK":
                    failed_to_fetch_ids.append(email_id)
                    continue
                emails.append(email_module.message_from_bytes(msg_data[0][1]))
                if set_flags:
                    server.store(email_id, '+FLAGS', set_flags)
                if del_flags:
                    server.store(email_id, '-FLAGS', del_flags)
            return emails, failed_to_fetch_ids

    async def list_mailboxes_async(self) -> list[str]:
        """
        List all mailboxes in the IMAP account. (Async)
        """
        def _sync():
            return self.list_mailboxes()

        return await asyncio.to_thread(_sync)

    async def get_emails_async(
        self,
        mailbox: str = "INBOX",
        search_criteria: str = "ALL",
        set_flags: str | None = "\\Seen",
        del_flags: str | None = None,
        max: int | None = None
    ) -> tuple[list[EmailMessage], list[bytes]]:
        """
        Retrieve emails from the specified mailbox
        matching the search criteria. (async)
        """
        def _sync():
            return self.get_emails(
                mailbox=mailbox,
                criteria=search_criteria,
                set_flags=set_flags,
                del_flags=del_flags,
                max=max
            )

        return await asyncio.to_thread(_sync)


class EmailManager:
    """
    Default email manager class that combines both EmailSender and EmailReader
    """
    def __init__(
            self,
            email: str,
            password: str,
            name: str | None = None,
            reply_to_email: str | None = None,
            reply_to_name: str | None = None,
            smtp_server: str | None = None,
            smtp_port: int | None = None,
            imap_server: str | None = None,
            imap_port: int | None = None,
            auth_for_sending: bool = False):
        self.email_sender = EmailSender(
            sender_email=email,
            sender_name=name,
            reply_to_email=reply_to_email,
            reply_to_name=reply_to_name,
            smtp_server=smtp_server,
            smtp_port=smtp_port,
            sender_password=password if auth_for_sending else None,
        )

        self.email_reader = EmailReader(
            email=email,
            password=password,
            imap_server=imap_server,
            imap_port=imap_port
        )

    def get_mailboxes(self, *args, **kwargs):
        return self.email_reader.list_mailboxes(*args, **kwargs)

    def get_emails(self, *args, **kwargs):
        return self.email_reader.get_emails(*args, **kwargs)

    def send_email(self, *args, **kwargs):
        return self.email_sender.send_email(*args, **kwargs)

    async def get_mailboxes_async(self, *args, **kwargs):
        return await self.email_reader.list_mailboxes_async(*args, **kwargs)

    async def get_emails_async(self, *args, **kwargs):
        return await self.email_reader.get_emails_async(*args, **kwargs)

    async def send_email_async(self, *args, **kwargs):
        return await self.email_sender.send_email_async(*args, **kwargs)



# Export main classes for easier imports
from .database_manager import DatabaseManager  # noqa: F401
from .token_session import ManagedOAuth2Session  # noqa: F401
from .email_handling import EmailManager  # noqa: F401
from .email_handling import EmailSender  # noqa: F401
from .email_handling import EmailReader  # noqa: F401



import pytest
import time
import types
from requests_oauthlib import OAuth2Session

from rkdigi import ManagedOAuth2Session

DUMMY_TOKEN_URL = "https://example.com/token"
DUMMY_CLIENT_ID = "dummy-client-id"
DUMMY_CLIENT_SECRET = "dummy-secret"


@pytest.fixture
def session():
    return ManagedOAuth2Session(
        token_url=DUMMY_TOKEN_URL,
        client_id=DUMMY_CLIENT_ID,
        client_secret=DUMMY_CLIENT_SECRET
    )


class DummyRequest:
    url = "http://example.com"
    headers = {}
    body = b""


class DummyResponse:
    status_code = 200
    request = DummyRequest()
    headers = {}
    text = ""


def test_managed_oauth2_session_init(session):
    """Test init"""
    assert session.token_url == DUMMY_TOKEN_URL
    assert session.client_id == DUMMY_CLIENT_ID
    assert session.client_secret == DUMMY_CLIENT_SECRET
    assert isinstance(session.extra_params, dict)
    # Properties
    assert session.token == {}
    assert session.access_token is None
    assert session.refresh_token_value is None
    assert session.access_token_expiry == 0


def test_init_raises_on_grant_type():
    """
    Test that specifying grant_type
    in extra_params raises ValueError
    """
    with pytest.raises(ValueError):
        ManagedOAuth2Session(
            token_url=DUMMY_TOKEN_URL,
            client_id=DUMMY_CLIENT_ID,
            client_secret=DUMMY_CLIENT_SECRET,
            extra_params={"grant_type": "client_credentials"}
        )


def test_reacquire_if_expired_fetching_token(session):
    session.token = {}
    session._fetching_token = True
    called = {}

    def fake_acquire_token():
        called["called"] = True

    session._acquire_token = types.MethodType(fake_acquire_token, session)
    session._reacquire_if_expired()
    # Should not call _acquire_token if _fetching_token is True
    assert not called
    session._fetching_token = False  # reset for other tests


def test_reacquire_if_expired_with_refresh_token(session, monkeypatch):
    session.token = {"refresh_token": "refresh",
                     "expires_at": time.time() - 10}
    called = {}

    def fake_acquire_token():
        called["called"] = True

    monkeypatch.setattr(session, "_acquire_token", fake_acquire_token)
    session._reacquire_if_expired()
    # Should not call _acquire_token if refresh_token is present
    assert not called


def test_reacquire_if_expired_token_not_expired(session, monkeypatch):
    session.token = {"expires_at": time.time() + 100}
    called = {}

    def fake_acquire_token():
        called["called"] = True

    monkeypatch.setattr(session, "_acquire_token", fake_acquire_token)
    session._reacquire_if_expired()
    # Should not call _acquire_token if token is not expired
    assert not called


def test_request_sets_auto_refresh(session, monkeypatch):
    session.token = {"refresh_token": "refresh",
                     "expires_at": time.time() + 100}
    called = {}

    def fake_super_request(self, method, url, **kwargs):
        called["called"] = True
        return "ok"

    monkeypatch.setattr(OAuth2Session, "request", fake_super_request)
    result = session.request("GET", "http://example.com")
    assert result == "ok"
    assert session.auto_refresh_url == DUMMY_TOKEN_URL


def test_request_no_refresh_token(session, monkeypatch):
    session.token = {}  # No refresh_token
    called = {}

    def fake_super_request(self, method, url, **kwargs):
        called["called"] = True
        return DummyResponse()

    def fake_fetch_token(*a, **kw):
        return {"access_token": "dummy", "expires_at": time.time() + 100}

    monkeypatch.setattr(OAuth2Session, "request", fake_super_request)
    monkeypatch.setattr(OAuth2Session, "fetch_token", fake_fetch_token)
    result = session.request("GET", "http://example.com")
    assert isinstance(result, DummyResponse)



import os
from urllib import parse
from typing import Literal
from contextlib import asynccontextmanager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session, DeclarativeMeta


class DatabaseManager:
    """
    A singleton class to manage database connections and sessions.
    Can handle PostgreSQL and MSSQL databases, with support for
    both synchronous and asynchronous modes.
    For synchronous PostgreSQL, it can also create tables
    based on a provided SQLAlchemy declarative base model.
    """
    _instances = {}

    def __new__(cls, profile_name: str, *args, **kwargs):
        # Ensure singleton behavior based on profile_name
        if profile_name in cls._instances:
            return cls._instances[profile_name]
        instance = super().__new__(cls)
        cls._instances[profile_name] = instance
        return instance

    def __init__(
        self,
        profile_name: str,
        db_type: Literal["postgres", "mssql"],
        airflow_connection_id: str | None = None,
        database: str | None = None,
        username: str | None = None,
        password: str | None = None,
        host: str | None = None,
        port: int | None = None,
        async_mode: bool = False,
        base_model: DeclarativeMeta | None = None
    ):
        # Checking if already initialized
        # to prevent re-initialization in singleton pattern.
        # returns object with the same profile_name without re-initializing it.
        # For re-initialization, dispose() must be called first.
        if getattr(self, "_initialized", False):
            return
        self._initialized = True

        self._async_mode = async_mode

        if self._async_mode:
            # import libs needed for asynchronous mode
            from sqlalchemy.ext.asyncio import create_async_engine, \
                async_sessionmaker, AsyncSession

        if db_type not in ("postgres", "mssql"):
            raise ValueError(
                f"db_type must be 'postgres' or 'mssql', got '{db_type}'"
            )
        elif airflow_connection_id:
            if self._async_mode:
                # Get connection string from Airflow Connection
                from airflow.hooks.base import BaseHook
                conn = BaseHook.get_connection(airflow_connection_id)
                conn_str = conn.get_uri()
                # default drivers in airflow is sync ones, so we need to
                # replace them with async drivers
                if db_type == "postgres":
                    driver = "postgresql+asyncpg"
                elif db_type == "mssql":
                    driver = "mssql+aioodbc"
                conn_str = f"{driver}://{conn_str.split('://', 1)[1]}"
                self._engine = create_async_engine(conn_str)
            else:
                # Getting SQLAlchemy engine from Airflow Hook (sync)
                if db_type == "postgres":
                    from airflow.providers.postgres.hooks.postgres \
                        import PostgresHook
                    hook = PostgresHook(postgres_conn_id=airflow_connection_id)
                elif db_type == "mssql":
                    from airflow.providers.microsoft.mssql.hooks.mssql \
                        import MsSqlHook
                    hook = MsSqlHook(mssql_conn_id=airflow_connection_id)
                self._engine = hook.get_sqlalchemy_engine()
        elif all(p is None for p in (username, password, host)):
            # user, pass and host are all None, try to get them from env vars
            # If also missing from there an error
            # will be raised later when trying to connect
            env_prefix = profile_name.upper() + "_"
            username = (os.getenv(env_prefix + "USERNAME")
                        or os.getenv(env_prefix + "USER"))
            password = (os.getenv(env_prefix + "PASSWORD")
                        or os.getenv(env_prefix + "PASS"))
            host = os.getenv(env_prefix + "HOST")
            database = os.getenv(env_prefix + "DATABASE")
            port_str = os.getenv(env_prefix + "PORT")
            port = int(port_str) if port_str is not None else None
            if not all((username, password, host)):
                raise ValueError(
                    "Missing environment variables "
                    f"for profile '{profile_name}'."
                )
        elif all((username, password, host)):
            # all required parameters are provided
            pass
        else:
            raise ValueError(
                "Insufficient parameters provided. Either provide "
                "`airflow_connection_id`, or provide all of "
                "`username`, `password`, and `host` (with optional "
                "`database` and `port`), or omit these arguments and "
                "set the corresponding environment variables derived "
                "from `profile_name` (e.g. '<PROFILE>_USERNAME', "
                "'<PROFILE>_PASSWORD', '<PROFILE>_HOST')."
            )

        # If airflow connection_id was provided, engine is already created
        # so skip the following part
        if not hasattr(self, "_engine"):
            # set driver based on db_type and if sync or async mode
            if db_type == "postgres":
                driver = "postgresql+asyncpg" if self._async_mode \
                    else "postgresql+psycopg2"
            elif db_type == "mssql":
                driver = "mssql+aioodbc" if self._async_mode \
                    else "mssql+pymssql"

            # Basic connection string
            conn_str = (
                f"{driver}://{parse.quote_plus(username)}:"
                f"{parse.quote_plus(password)}@{parse.quote_plus(host)}"
            )

            # Add port and database if provided
            if port:
                conn_str += f":{port}"
            if database:
                conn_str += f"/{parse.quote_plus(database)}"

            if not self._async_mode:
                self._engine = create_engine(conn_str)
            else:
                # 'aioodbc' depends on ODBC driver and it must be specified.
                if db_type == "mssql":
                    import pyodbc
                    # Check system for available ODBC drivers
                    # and add the first suitable one to connection string
                    drivers = [d for d in pyodbc.drivers()
                               if "ODBC Driver" in d and "SQL Server" in d]
                    conn_str += f"?driver={drivers[0].replace(' ', '+')}"
                self._engine = create_async_engine(conn_str)

        if not self._async_mode:
            self.can_connect()  # Check connection on init
            # Create tables if base_model is provided.
            if base_model:
                base_model.metadata.create_all(self._engine)
            self._Session = sessionmaker(bind=self._engine)
        else:
            # Check connection is skipped on init for async mode
            # Can not call the async method from sync __init__
            self._Session = async_sessionmaker(
                bind=self._engine,
                class_=AsyncSession
            )

    def can_connect(self) -> bool | None:
        """Check if a connection can be established."""
        if self._async_mode:
            raise RuntimeError("Use can_connect_async for async mode.")
        with self._engine.connect():
            pass
        return True

    def get_session(self) -> Session:
        """Get a synchronous database session."""
        if self._async_mode:
            raise RuntimeError("In async mode, use get_session_async instead.")
        return self._Session()

    def dispose(self) -> None:
        """Dispose of the database engine and clear the instance."""
        if hasattr(self, "_engine") and not self._async_mode:
            self._engine.dispose()
            self._engine = None
            self._Session = None
            self._initialized = False
            for name, inst in list(self._instances.items()):
                if inst is self:
                    del self._instances[name]
        else:
            raise RuntimeError("In async mode, use dispose_async instead.")

    async def can_connect_async(self) -> bool | None:
        """Check if a connection can be established asynchronously."""
        if not self._async_mode:
            raise RuntimeError("Use can_connect for sync mode.")
        async with self._engine.connect():
            pass
        return True

    @asynccontextmanager
    async def get_session_async(self):
        """Get an asynchronous database session."""
        if not self._async_mode:
            raise RuntimeError("Not in async mode.")
        async with self._Session() as session:
            yield session

    async def dispose_async(self):
        """Dispose of the async database engine and clear the instance."""
        if hasattr(self, "_engine") and self._async_mode:
            await self._engine.dispose()
            self._engine = None
            self._Session = None
            self._initialized = False
            for name, inst in list(self._instances.items()):
                if inst is self:
                    del self._instances[name]
        else:
            raise RuntimeError("Not in async mode.")

    async def create_tables_async(self, base_model: DeclarativeMeta):
        """
        Create tables based on the provided
        SQLAlchemy declarative base model.
        """
        if not self._async_mode:
            raise RuntimeError("Not in async mode.")
        async with self._engine.begin() as conn:
            await conn.run_sync(base_model.metadata.create_all)



import sys
import types
import pytest
import asyncio
import sqlalchemy
from rkdigi import DatabaseManager

# Mock psycopg2
sys.modules["psycopg2"] = types.ModuleType("psycopg2")
sys.modules["psycopg2.extensions"] = types.ModuleType("psycopg2.extensions")
sys.modules["psycopg2.extras"] = types.ModuleType("psycopg2.extras")
sys.modules["psycopg2"].paramstyle = "pyformat"

# Mock pymssql
sys.modules["pymssql"] = types.ModuleType("pymssql")
sys.modules["pymssql"].paramstyle = "pyformat"
sys.modules["pymssql"].__version__ = "2.2.0"

# Mock airflow and airflow providers
sys.modules["airflow.hooks.base"] = types.ModuleType("airflow.hooks.base")
sys.modules["airflow.providers.postgres.hooks.postgres"] = \
    types.ModuleType("airflow.providers.postgres.hooks.postgres")
sys.modules["airflow.providers.microsoft.mssql.hooks.mssql"] = \
    types.ModuleType("airflow.providers.microsoft.mssql.hooks.mssql")


@pytest.fixture(autouse=True)
def clear_db_manager_instances():
    DatabaseManager._instances.clear()


# General Tests
def test_singleton_behavior(monkeypatch):
    monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)
    manager1 = DatabaseManager(
        profile_name="singleton_test",
        db_type="postgres",
        username="user",
        password="pass",
        host="localhost",
    )
    manager2 = DatabaseManager(
        profile_name="singleton_test",
        db_type="postgres",
        username="user2",
        password="pass2",
        host="localhost",
    )
    assert manager1 is manager2
    manager1.dispose()
    assert "singleton_test" not in DatabaseManager._instances


def test_insufficient_credentials():
    with pytest.raises(ValueError):
        DatabaseManager(
            profile_name="incomplete",
            db_type="postgres",
            username="user"
        )


def test_invalid_db_type():
    with pytest.raises(ValueError):
        DatabaseManager(
            profile_name="badtype",
            db_type="sqlite"
        )


# Sync Tests
@pytest.mark.parametrize("db_type", ["postgres", "mssql"])
def test_sync_manager_init_airflow(db_type, monkeypatch):
    # Lines 80->84 in database_manager.py are covered by this test,
    # but due to monkeypatching/pytest-cov limitations
    # it is not visible in coverage report.
    class DummyEngine:
        def dispose(self):
            pass

    class DummyHook:
        def get_sqlalchemy_engine(self):
            return DummyEngine()

    sys.modules["airflow.providers.postgres.hooks.postgres"].PostgresHook = \
        lambda postgres_conn_id: DummyHook()
    sys.modules["airflow.providers.microsoft.mssql.hooks.mssql"].MsSqlHook = \
        lambda mssql_conn_id: DummyHook()
    monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)
    manager = DatabaseManager(
        profile_name="airflowtest",
        db_type=db_type,
        airflow_connection_id="dummy_id"
    )
    assert isinstance(manager, DatabaseManager)


@pytest.mark.parametrize("db_type", ["postgres", "mssql"])
def test_sync_manager_init_creds(db_type, monkeypatch):
    # Lines 124->129 in database_manager.py are covered by this test,
    # but due to monkeypatching/pytest-cov limitations
    # it is not visible in coverage report.
    monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)
    manager = DatabaseManager(
        profile_name="testprofile",
        db_type=db_type,
        username="user",
        password="pass",
        host="localhost",
        database="somedb",
        port=1234
    )
    assert isinstance(manager, DatabaseManager)


@pytest.mark.parametrize("db_type", ["postgres", "mssql"])
def test_sync_manager_init_env_vars(db_type, monkeypatch):
    monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)
    monkeypatch.setenv("TESTDB_USERNAME", "envuser")
    monkeypatch.setenv("TESTDB_PASSWORD", "envpass")
    monkeypatch.setenv("TESTDB_HOST", "envhost")
    monkeypatch.setenv("TESTDB_PORT", "1234")
    monkeypatch.setenv("TESTDB_DATABASE", "envdb")
    manager = DatabaseManager(
        profile_name="testdb",
        db_type=db_type
    )
    assert manager._initialized


def test_sync_manager_init_env_vars_missing(monkeypatch):
    monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)
    monkeypatch.setenv("TESTDB_USERNAME", "envuser")
    monkeypatch.setenv("TESTDB_PASSWORD", "envpass")
    with pytest.raises(ValueError):
        DatabaseManager(
            profile_name="testdb",
            db_type='postgres'
        )


def test_sync_manager_init_with_base_model(monkeypatch):
    Base = sqlalchemy.orm.declarative_base()

    class DummyEngine:
        def dispose(self): pass

    monkeypatch.setattr(
        "rkdigi.database_manager.create_engine",
        lambda *a, **k: DummyEngine()
    )
    monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)

    Base.create_all_called = False

    def dummy_create_all(engine):
        Base.create_all_called = True
        assert isinstance(engine, DummyEngine)
    Base.metadata.create_all = dummy_create_all

    manager = DatabaseManager(
        profile_name="basemodel_test",
        db_type="postgres",
        username="user",
        password="pass",
        host="localhost",
        base_model=Base
    )
    assert isinstance(manager, DatabaseManager)
    assert Base.create_all_called


def test_can_connect_sync(monkeypatch):
    class DummyConnection:
        def __enter__(self): return self
        def __exit__(self, exc_type, exc_val, exc_tb): pass

    class DummyEngine:
        def connect(self): return DummyConnection()

    monkeypatch.setattr(sqlalchemy,
                        "create_engine",
                        lambda *a, **k: DummyEngine())
    monkeypatch.setattr("sqlalchemy.orm.sessionmaker",
                        lambda *a, **k: lambda: None)

    # Mock can_connect during init and restore afterwards
    orig_can_connect = DatabaseManager.can_connect
    try:
        monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)
        manager = DatabaseManager(
            profile_name="connect_test",
            db_type="postgres",
            username="user",
            password="pass",
            host="localhost",
        )
    finally:
        DatabaseManager.can_connect = orig_can_connect

    monkeypatch.setattr(manager, "_engine", DummyEngine())
    assert manager.can_connect() is True
    manager._async_mode = True
    with pytest.raises(RuntimeError):
        manager.can_connect()


def test_get_session_sync(monkeypatch):
    monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)
    manager = DatabaseManager(
        profile_name="session_test",
        db_type="postgres",
        username="user",
        password="pass",
        host="localhost",
    )

    class DummySession:
        pass
    manager._Session = lambda: DummySession()
    session = manager.get_session()
    assert isinstance(session, DummySession)
    with pytest.raises(RuntimeError):
        manager._async_mode = True
        manager.get_session()


def test_dispose_sync(monkeypatch):
    monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)
    manager = DatabaseManager(
        profile_name="dispose_test",
        db_type="postgres",
        username="user",
        password="pass",
        host="localhost",
    )

    assert "dispose_test" in DatabaseManager._instances
    manager.dispose()
    assert "dispose_test" not in DatabaseManager._instances
    assert manager._engine is None
    assert manager._Session is None
    assert manager._initialized is False
    with pytest.raises(RuntimeError):
        manager._async_mode = True
        manager.dispose()


# Async Tests
# Mock sqlalchemy async
class DummyAsyncConnection:
    async def __aenter__(self): return self
    async def __aexit__(self, exc_type, exc_val, exc_tb): pass


class DummyAsyncEngine:
    def connect(self): return DummyAsyncConnection()
    async def dispose(self): pass


def dummy_create_async_engine(conn_str):
    return DummyAsyncEngine()


def dummy_async_sessionmaker(*args, **kwargs):
    class DummySession:
        async def __aenter__(self): return self
        async def __aexit__(self, exc_type, exc, tb): pass
        async def close(self): pass
    return DummySession


sys.modules["sqlalchemy.ext.asyncio"] = \
    types.ModuleType("sqlalchemy.ext.asyncio")
sys.modules["sqlalchemy.ext.asyncio"].create_async_engine = \
    dummy_create_async_engine
sys.modules["sqlalchemy.ext.asyncio"].async_sessionmaker = \
    dummy_async_sessionmaker
sys.modules["sqlalchemy.ext.asyncio"].AsyncSession = None

# Mock pyodbc for async mssql
pyodbc = types.ModuleType("pyodbc")
pyodbc.drivers = lambda: ["ODBC Driver 18 for SQL Server"]
sys.modules["pyodbc"] = pyodbc


@pytest.mark.asyncio
@pytest.mark.parametrize("db_type", ["postgres", "mssql"])
async def test_async_manager_init_airflow(db_type, monkeypatch):
    # Lines 70->72 in database_manager.py are covered by this test,
    # but due to monkeypatching/pytest-cov limitations
    # it is not visible in coverage report.
    class DummyConnection:
        def get_uri(self):
            return "dummy://user:pass@host:1234/dbname"

    class DummyBaseHook:
        @staticmethod
        def get_connection(conn_id):
            return DummyConnection()

    sys.modules["airflow.hooks.base"].BaseHook = DummyBaseHook

    manager = DatabaseManager(
        profile_name=f"airflowtest_async_{db_type}",
        db_type=db_type,
        airflow_connection_id="dummy_id",
        async_mode=True
    )
    assert isinstance(manager, DatabaseManager)


@pytest.mark.asyncio
@pytest.mark.parametrize("db_type", ["postgres", "mssql"])
async def test_async_manager_init_creds(db_type, monkeypatch):
    manager = DatabaseManager(
        profile_name=f"airflowtest_async_{db_type}",
        db_type=db_type,
        username="user",
        password="pass",
        host="localhost",
        async_mode=True
    )
    assert isinstance(manager, DatabaseManager)


def test_can_connect_async():
    manager = DatabaseManager(
        profile_name="connect_async_test",
        db_type="postgres",
        username="user",
        password="pass",
        host="localhost",
        async_mode=True
    )
    manager._engine = DummyAsyncEngine()

    assert asyncio.run(manager.can_connect_async()) is True
    with pytest.raises(RuntimeError):
        manager._async_mode = False
        asyncio.run(manager.can_connect_async())


def test_get_session_async(monkeypatch):
    monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)

    class DummyAsyncSession:
        async def __aenter__(self): return self
        async def __aexit__(self, exc_type, exc_val, exc_tb): pass
        async def close(self): pass

    manager = DatabaseManager(
        profile_name="session_async_test",
        db_type="postgres",
        username="user",
        password="pass",
        host="localhost",
        async_mode=True
    )
    manager._Session = lambda: DummyAsyncSession()

    async def run():
        async with manager.get_session_async() as session:
            assert isinstance(session, DummyAsyncSession)
        with pytest.raises(RuntimeError):
            manager._async_mode = False
            async with manager.get_session_async() as _:
                pass
    asyncio.run(run())


def test_dispose_async(monkeypatch):
    monkeypatch.setattr(DatabaseManager, "can_connect", lambda self: True)
    manager = DatabaseManager(
        profile_name="dispose_async_test",
        db_type="postgres",
        username="user",
        password="pass",
        host="localhost",
        async_mode=True
    )
    manager._engine = DummyAsyncEngine()

    assert "dispose_async_test" in DatabaseManager._instances
    assert manager is DatabaseManager._instances["dispose_async_test"]

    async def run():
        await manager.dispose_async()
        assert "dispose_async_test" not in DatabaseManager._instances
        assert manager._engine is None
        assert manager._Session is None
        assert manager._initialized is False
        with pytest.raises(RuntimeError):
            manager._async_mode = False
            await manager.dispose_async()
    asyncio.run(run())


@pytest.mark.asyncio
async def test_create_tables_async(monkeypatch):
    class DummyAsyncConn:
        async def __aenter__(self): return self
        async def __aexit__(self, exc_type, exc_val, exc_tb): pass

        async def run_sync(self, fn):
            DummyBaseModel.create_all_called = True
            return fn(DummyBaseModel)

    class DummyAsyncEngine:
        def begin(self):
            return DummyAsyncConn()

    class DummyBaseModel:
        metadata = types.SimpleNamespace(create_all=lambda engine: None)
        create_all_called = False

    manager = DatabaseManager(
        profile_name="async_create_tables_test",
        db_type="postgres",
        username="user",
        password="pass",
        host="localhost",
        async_mode=True
    )
    manager._engine = DummyAsyncEngine()
    DummyBaseModel.create_all_called = False
    await manager.create_tables_async(DummyBaseModel)
    assert DummyBaseModel.create_all_called

    with pytest.raises(RuntimeError):
        manager._async_mode = False
        await manager.create_tables_async(DummyBaseModel)
